@@ -2,10 +2,10 @@
 PythonSum = sum
 PythonAny = any
 PythonMax = max
-from numpy import inf, asfarray, copy, all, any, atleast_2d, zeros, dot, asarray, atleast_1d, \
+from numpy import inf, asfarray, all, any, atleast_2d, zeros, dot, asarray, atleast_1d, \
 ones, ndarray, where, array, nan, vstack, eye, array_equal, isscalar, log, hstack, sum as npSum, prod, nonzero,\
-isnan, asscalar, zeros_like, ones_like, logical_and, logical_or, isinf, logical_not, logical_xor, \
-tile, float64, searchsorted, int8, int16, int32, int64, isfinite, log2, string_, asanyarray, bool_
+asscalar, zeros_like, logical_and, \
+tile, searchsorted, int8, int16, int32, int64, string_, asanyarray
 
 import operator, numpy as np
 
@@ -15,66 +15,32 @@ try:
 except ImportError:
     from numpy import nanmin, nanmax
 
-from expression import getitem_expression, add_expression, mul_expression
+from expression import getitem_expression, add_expression, mul_expression, neg_expression, div_expression, \
+rdiv_expression, pow_expression, rpow_expression
+from iqg import iqg
+
 from boundsurf import boundsurf
-try:
-    from boundsurf2 import boundsurf2
-except ImportError:
-    boundsurf2 = boundsurf
+from boundsurf2 import boundsurf2
     
 from FDmisc import FuncDesignerException, Diag, Eye, pWarn, scipyAbsentMsg, scipyInstalled, \
-raise_except, DiagonalType, isPyPy, formResolveSchedule
+raise_except, DiagonalType, isPyPy, formResolveSchedule, Len, SP_eye, SparseMatrixConstructor, \
+isspmatrix, Hstack
+
 from ooPoint import ooPoint
 from FuncDesigner.multiarray import multiarray
-from Interval import Interval, adjust_lx_WithDiscreteDomain, adjust_ux_WithDiscreteDomain, mul_interval,\
-pow_const_interval, pow_oofun_interval, div_interval, add_interval, add_const_interval, \
-neg_interval, defaultIntervalEngine#, rdiv_interval
+from derivativeMisc import getDerivativeSelf, mul_aux_d, _D
+from Interval import Interval, mul_interval, pow_const_interval, pow_oofun_interval, div_interval, \
+add_interval, add_const_interval, neg_interval, defaultIntervalEngine#, rdiv_interval
 #import inspect
 from baseClasses import OOArray, Stochastic, OOFun
-from logic import AND, NOT, EQUIVALENT, nlh_not
-
-Copy = lambda arg: asscalar(arg) if type(arg)==ndarray and arg.size == 1 else arg.copy() if hasattr(arg, 'copy') else copy(arg)
-Len = lambda x: 1 if isscalar(x) else x.size if type(x)==ndarray else x.values.size if isinstance(x, Stochastic) else len(x)
 
 try:
-    from DerApproximator import get_d1, check_d1
+    from DerApproximator import check_d1
     DerApproximatorIsInstalled = True
 except:
     DerApproximatorIsInstalled = False
 
-try:
-    import scipy
-    #from scipy import sparse
-    from scipy.sparse import hstack as HstackSP, vstack as VstackSP, isspmatrix_csc, isspmatrix_csr, eye as SP_eye, lil_matrix as SparseMatrixConstructor
-    def Hstack(Tuple):
-        ind = where([isscalar(elem) or prod(elem.shape)!=0 for elem in Tuple])[0].tolist()
-        elems = [Tuple[i] for i in ind]
-        if any([isspmatrix(elem) for elem in elems]):
-            return HstackSP(elems)
-        
-        s = set([(0 if isscalar(elem) else elem.ndim) for elem in elems])
-        ndim = max(s)
-        if ndim <= 1:  return hstack(elems)
-        assert ndim <= 2 and 1 not in s, 'bug in FuncDesigner kernel, inform developers'
-        return hstack(elems) if 0 not in s else hstack([atleast_2d(elem) for elem in elems])
-    def Vstack(Tuple):
-        ind = where([isscalar(elem) or prod(elem.shape)!=0 for elem in Tuple])[0].tolist()
-        elems = [Tuple[i] for i in ind]
-        if any([isspmatrix(elem) for elem in elems]):
-            return VstackSP(elems)
-        else:
-            return vstack(elems)
-#        s = set([(0 if isscalar(elem) else elem.ndim) for elem in elems])
-#        ndim = max(s)
-#        if ndim <= 1:  return hstack(elems)
-#        assert ndim <= 2 and 1 not in s, 'bug in FuncDesigner kernel, inform developers'
-#        return hstack(elems) if 0 not in s else hstack([atleast_2d(elem) for elem in elems])        
-    from scipy.sparse import isspmatrix
-except:
-    scipy = None
-    isspmatrix = lambda *args,  **kwargs:  False
-    Hstack = hstack
-    Vstack = vstack
+
 
 class oofun(OOFun):
     #TODO:
@@ -355,158 +321,10 @@ class oofun(OOFun):
 
         return r            
     
-    def iqg(self, domain, dtype = float, lb=None, ub=None, UB = None):
-        if type(domain) != ooPoint:
-            domain = ooPoint(domain, skipArrayCast=True)
-            domain.isMultiPoint=True
-        domain.useSave = True
-        r0 = self.interval(domain, dtype, resetStoredIntervals = False)
-        
-        r0.lb, r0.ub = atleast_1d(r0.lb).copy(), atleast_1d(r0.ub).copy() # is copy required?
-        
-        # TODO: get rid of useSave
-        domain.useSave = False
-        
-        # TODO: rework it with indexation of required data
-        if lb is not None and ub is not None:
-            ind = logical_or(logical_or(r0.ub < lb, r0.lb > ub), all(logical_and(r0.lb >= lb, r0.ub <= ub)))
-        elif UB is not None:
-            ind = r0.lb > UB
-        else:
-            ind = None
-        
-        useSlicing = False
-        
-        if ind is not None:
-            if all(ind):
-                return {}, r0
-            j = where(~ind)[0]
-            #DOESN'T WORK FOR FIXED OOVARS AND DefiniteRange != TRUE YET
-            if 0 and j.size < 0.85*ind.size:  # at least 15% of values to skip
-                useSlicing = True
-                tmp = []
-                for key, val in domain.storedIntervals.items():
-                    Interval, definiteRange = val
-                    if type(definiteRange) not in (bool, bool_):
-                        definiteRange = definiteRange[j]
-                    tmp.append((key, (Interval[:, j], definiteRange)))
-                _storedIntervals = dict(tmp)
-                
-                Tmp = []
-                for key, val in domain.storedSums.items():
-                    # TODO: rework it
-                    R0, DefiniteRange0 = val.pop(-1)
-                    #R0, DefiniteRange0 = val[-1]
-                    R0 = R0[:, j]
-                    if type(DefiniteRange0) not in (bool, bool_):
-                        DefiniteRange0 = DefiniteRange0[j]
-                    tmp = []
-                    for k,v in val.items():
-                        # TODO: rework it
-#                        if k is (-1): continue
-                        v = v[:, j]
-                        tmp.append((k,v))
-                    val = dict(tmp)
-                    val[-1] = (R0, DefiniteRange0)
-                    Tmp.append((key,val))
-                _storedSums = dict(Tmp)
-                #domain.storedSums = dict(tmp)
-                
-                Tmp = []
-                for key, val in domain.items():
-                    lb_,ub_ = val
-                    # TODO: rework it when lb, ub will be implemented as 2-dimensional
-                    Tmp.append((key, (lb_[j],ub_[j])))
-                dictOfFixedFuncs = domain.dictOfFixedFuncs
-                domain2 = ooPoint(Tmp, skipArrayCast=True)
-                domain2.storedSums = _storedSums
-                domain2.storedIntervals = _storedIntervals
-                domain2.dictOfFixedFuncs = dictOfFixedFuncs
-                domain2.isMultiPoint=True
-                domain = domain2
-                
-        domain.useAsMutable = True
-        
-        r = {}
-        Dep = (self._getDep() if not self.is_oovar else set([self])).intersection(domain.keys())
-        
-        for i, v in enumerate(Dep):
-            domain.modificationVar = v
-            r_l, r_u = self._iqg(domain, dtype, r0)
-            if useSlicing and r_l is not r0:# r_l is r0 when array_equal(lb, ub)
-                lf1, lf2, uf1, uf2 = r_l.lb, r_u.lb, r_l.ub, r_u.ub
-                Lf1, Lf2, Uf1, Uf2 = Copy(r0.lb), Copy(r0.lb), Copy(r0.ub), Copy(r0.ub)
-                Lf1[:, j], Lf2[:, j], Uf1[:, j], Uf2[:, j] = lf1, lf2, uf1, uf2
-                r_l.lb, r_u.lb, r_l.ub, r_u.ub = Lf1, Lf2, Uf1, Uf2
-                if type(r0.definiteRange) not in (bool, bool_):
-                    d1, d2 = r_l.definiteRange, r_u.definiteRange
-                    D1, D2 = atleast_1d(r0.definiteRange).copy(), atleast_1d(r0.definiteRange).copy()
-                    D1[j], D2[j] = d1, d2
-                    r_l.definiteRange, r_u.definiteRange = D1, D2
-                
-            r[v] = r_l, r_u
-            if not self.isUncycled:
-                lf1, lf2, uf1, uf2 = r_l.lb, r_u.lb, r_l.ub, r_u.ub
-                lf, uf = nanmin(vstack((lf1, lf2)), 0), nanmax(vstack((uf1, uf2)), 0)
-                if i == 0:
-                    L, U = lf.copy(), uf.copy()
-                else:
-                    L[L<lf] = lf[L<lf].copy()
-                    U[U>uf] = uf[U>uf].copy()
-        if not self.isUncycled:
-            for R in r.values():
-                r1, r2 = R
-                if type(r1.lb) != ndarray:
-                    r1.lb, r2.lb, r1.ub, r2.ub = atleast_1d(r1.lb), atleast_1d(r2.lb), atleast_1d(r1.ub), atleast_1d(r2.ub)
-                r1.lb[r1.lb < L] = L[r1.lb < L]
-                r2.lb[r2.lb < L] = L[r2.lb < L]
-                r1.ub[r1.ub > U] = U[r1.ub > U]
-                r2.ub[r2.ub > U] = U[r2.ub > U]
-            
-            r0.lb[r0.lb < L] = L[r0.lb < L]
-            r0.ub[r0.ub > U] = U[r0.ub > U]
-            
-        # for more safety
-        domain.useSave = True
-        domain.useAsMutable = False
-        domain.modificationVar = None 
-        domain.storedIntervals = {}
-        
-        return r, r0
-    
-    def _iqg(self, domain, dtype, r0):
-        v = domain.modificationVar
-        v_0 = domain[v]
-        lb, ub = v_0[0], v_0[1]
+    iqg = iqg
 
-        if v.domain is not None and array_equal(lb, ub):
-            return r0,r0 
-
-        assert dtype in (float, float64, int32, int16),  'other types unimplemented yet'
-        middle = 0.5 * (lb+ub)
-        
-        if v.domain is not None:
-            middle1, middle2 = middle.copy(), middle.copy()
-            adjust_ux_WithDiscreteDomain(middle1, v)
-            adjust_lx_WithDiscreteDomain(middle2, v)
-        else:
-            middle1 = middle2 = middle
-        
-        domain[v] = (v_0[0], middle1)
-        domain.localStoredIntervals = {}
-        r_l = self.interval(domain, dtype, resetStoredIntervals = False)
-
-        domain[v] = (middle2, v_0[1])
-        domain.localStoredIntervals = {}
-        r_u = self.interval(domain, dtype, resetStoredIntervals = False)
-        
-        domain[v] = v_0
-        domain.localStoredIntervals = {}
-        return r_l, r_u
-            
     __pos__ = lambda self: self
 
-        
     # overload "a+b"
     # @checkSizes
     def __add__(self, other):
@@ -594,14 +412,7 @@ class oofun(OOFun):
         r.d = raise_except
         r.vectorized = True
         r._interval_ = lambda *args, **kw: neg_interval(self, *args, **kw)
-        def expression(*args, **kw):
-            r = self.expression(**kw)
-            needBrackets = '+' in r or '-' in r
-            if needBrackets:
-                return '- (' + r + ')'
-            return '- ' + r
-            
-        r.expression = expression#lambda *args, **kw: '-' + self.expression(**kw)
+        r.expression =  lambda *args, **kw: neg_expression(self, *args, **kw)
         return r
         
     # overload "a-b"
@@ -661,24 +472,13 @@ class oofun(OOFun):
 
 #            if other.size == 1 or 'size' in self.__dict__ and self.size in (1, other.size):
             if other.size == 1:
-                r._D = lambda *args, **kwargs: dict([(key, value/other) for key, value in self._D(*args, **kwargs).items()])
+                r._D = lambda *args, **kwargs: dict((key, value/other) for key, value in self._D(*args, **kwargs).items())
                 r.d = raise_except
             
         # r.discrete = self.discrete and (?)
         #r.isCostly = True
         r.vectorized = True
-        
-        def expression(*args, **kw):
-            r1 = self.expression(**kw)
-            needBrackets1 = '+' in r1 or '-' in r1  or '/' in r1#or '*' in r1
-            R1 = '(' + r1 + ')' if needBrackets1 else r1
-
-            r2 = other.expression(**kw) if isinstance(other, oofun) else  str(other)
-            needBrackets2 = '+' in r2 or '-' in r2 or '*' in r2 or '/' in r2
-            R2 = '(' + r2 + ')' if needBrackets2 else r2
-
-            return R1 + '/' + R2
-        r.expression = expression
+        r.expression = lambda *args, **kw: div_expression(self, other, *args, **kw)
         return r
 
     def __rdiv__(self, other):
@@ -703,14 +503,8 @@ class oofun(OOFun):
             return 0 if order == 0 else inf
         r.getOrder = getOrder
         r.vectorized = True
-        
-        def expression(*args, **kw):
-            r1 = self.expression(**kw)
-            needBrackets1 = '+' in r1 or '-' in r1  or '/' in r1 or '*' in r1
-            R1 = '(' + r1 + ')' if needBrackets1 else r1
 
-            return str(other) + '/' + R1
-        r.expression = expression
+        r.expression = lambda *args, **kw: rdiv_expression(self, other, *args, **kw)
         return r
 
     # overload "a*b"
@@ -771,7 +565,7 @@ class oofun(OOFun):
             r._getFuncCalcEngine = lambda *args,  **kwargs: other * self._getFuncCalcEngine(*args,  **kwargs)
 
             if isscalar(other) or asarray(other).size == 1:  # other may be array-like
-                r._D = lambda *args, **kwargs: dict([(key, value * other) for key, value in self._D(*args, **kwargs).items()])
+                r._D = lambda *args, **kwargs: dict((key, value * other) for key, value in self._D(*args, **kwargs).items())
                 r.d = raise_except
             else:
                 r.d = lambda x: mul_aux_d(x, other)
@@ -865,19 +659,8 @@ class oofun(OOFun):
         r.vectorized = True
         if other_is_oofun or not isInt:
             r._lower_domain_bound = 0.0
-        
-        def expression(*args, **kw):
-            r1 = self.expression(**kw)
-            needBrackets1 = '+' in r1 or '-' in r1 or '*' in r1 or '/' in r1 or '^' in r1
-            R1 = '(' + r1 + ')' if needBrackets1 else r1
-
-            r2 = other.expression(**kw) if isinstance(other, oofun) else str(other)
-            needBrackets2 = '+' in r2 or '-' in r2 or '*' in r2 or '/' in r2 or '^' in r2
-            R2 = '(' + r2 + ')' if needBrackets2 else r2
-            pow_symbol = kw.get('pow', '^') 
-            return R1 + pow_symbol + R2
             
-        r.expression = expression
+        r.expression = lambda *args, **kw: pow_expression(self, other, *args, **kw)
         return r
 
     def __rpow__(self, other):
@@ -908,13 +691,8 @@ class oofun(OOFun):
             return exp_b_interval(log(other) * lb_ub, r1, definiteRange, domain)
             
         r._interval_ = lambda *args, **kw: rpow_interval(r, other, *args, **kw)
-        def expression(*args, **kw):
-            r1 = self.expression(**kw)
-            needBrackets1 = '+' in r1 or '-' in r1  or '/' in r1 or '*' in r1 or '^' in r1
-            R1 = '(' + r1 + ')' if needBrackets1 else r1
-            pow_symbol = kw.get('pow', '^') 
-            return str(other) + pow_symbol + R1
-        r.expression = expression
+
+        r.expression = lambda *args, **kw: rpow_expression(self, other, *args, **kw)
         return r
 
     def __xor__(self, other): raise FuncDesignerException('For power of oofuns use a**b, not a^b')
@@ -1063,7 +841,7 @@ class oofun(OOFun):
         r.name = self.name + ' >= ' + Other
         def expression(*args, **kw):
             r1 = self.expression(**kw)
-            r2 = other.expression(**kw) if isinstance(other, oofun) else Other if kw.get('truncation', True) else str(other)
+            r2 = other.expression(**kw) if isinstance(other, oofun) else Other if kw.get('truncation', False) else str(other)
             return r1 + ' >= '  + r2
         r.expression = expression
         return r
@@ -1090,7 +868,7 @@ class oofun(OOFun):
         r.name = self.name + ' <= ' + Other
         def expression(*args, **kw):
             r1 = self.expression(**kw)
-            r2 = other.expression(**kw) if isinstance(other, oofun) else Other if kw.get('truncation', True) else str(other)
+            r2 = other.expression(**kw) if isinstance(other, oofun) else Other if kw.get('truncation', False) else str(other)
             return r1 + ' <= ' + r2
         r.expression = expression
         return r            
@@ -1139,7 +917,7 @@ class oofun(OOFun):
             
         def expression(*args, **kw):
             r1 = self.expression(**kw)
-            r2 = other.expression(**kw) if isinstance(other, oofun) else Other if kw.get('truncation', True) else str(other)
+            r2 = other.expression(**kw) if isinstance(other, oofun) else Other if kw.get('truncation', False) else str(other)
             return r1 + ' == ' + r2
         r.expression = expression
         
@@ -1204,7 +982,7 @@ class oofun(OOFun):
                 r_oofuns.append(tmp)
             r = set(r_oovars)
 
-            # Python 2.5 sel.update fails on empty input
+            # Python 2.5 set.update fails on empty input
             if len(r_oofuns)!=0: r.update(*r_oofuns)
             if len(r_oovars) + sum([len(elem) for elem in r_oofuns]) != len(r):
                 isUncycled = False
@@ -1373,7 +1151,7 @@ class oofun(OOFun):
 
             # TODO: rework it (for input with ooarays)
             try:
-                t1 = dict([(elem, (x if isinstance(x, dict) else x.xf)[elem]) for elem in dep]) if self.isCostly else None
+                t1 = dict((elem, (x if isinstance(x, dict) else x.xf)[elem]) for elem in dep) if self.isCostly else None
                 #t1 = dict([(elem, copy((x if isinstance(x, dict) else x.xf)[elem])) for elem in dep]) if self.isCostly else None
                 if t1 is not None:
                     t2 = tmp.copy() if isinstance(tmp, (ndarray, Stochastic)) else tmp
@@ -1445,300 +1223,17 @@ class oofun(OOFun):
             return rr if not is_oofun else rr[initialVars]
         else:
             raise FuncDesignerException('Incorrect argument resultKeysType, should be "vars" or "names"')
-            
-            
-    def _D(self, x, fixedVarsScheduleID, Vars=None, fixedVars = None, useSparse = 'auto'):
-        if self.is_oovar: 
-            if (fixedVars is not None and self in fixedVars) or (Vars is not None and self not in Vars):
-                return {} 
-            tmp = x[self]
-            return {self:Eye(asarray(tmp).size)} if not isinstance(tmp, multiarray) else {self: ones_like(tmp).view(multiarray)}
-            
-        if self.input[0] is None: return {} # fixed oofun. TODO: implement input = [] properly
-            
-        if self.discrete: 
-            return {}
-            #raise FuncDesignerException('The oofun or oovar instance has been declared as discrete, no derivative is available')
-        
-        CondSamePointByID = True if isinstance(x, ooPoint) and self._point_id1 == x._id else False
-        sameVarsScheduleID = fixedVarsScheduleID == self._lastDiffVarsID 
-        
-        dep = self._getDep()
-        
-        rebuildFixedCheck = not sameVarsScheduleID
-        if rebuildFixedCheck:
-            self._isFixed = (fixedVars is not None and dep.issubset(fixedVars)) or (Vars is not None and dep.isdisjoint(Vars))
-        if self._isFixed: return {}
-        ##########################
-        
-        # TODO: optimize it. Omit it for simple cases.
-        #isTransmit = self._usedIn == 1 # Exactly 1! not 0, 2, ,3, 4, etc
-        #involveStore = not isTransmit or self._directlyDwasInwolved
-        involveStore = self.isCostly
 
-        #cond_same_point = hasattr(self, '_d_key_prev') and sameDerivativeVariables and (CondSamePointByID or (involveStore and         all([array_equal(x[elem], self.d_key_prev[elem]) for elem in dep])))
-        
-        cond_same_point = sameVarsScheduleID and \
-        ((CondSamePointByID and self._d_val_prev is not None) or \
-        (involveStore and self._d_key_prev is not None and all([array_equal(x[elem], self._d_key_prev[elem]) for elem in dep])))
-        
-        if cond_same_point:
-            self.same_d += 1
-            #return deepcopy(self.d_val_prev)
-            return dict((key, Copy(val)) for key, val in self._d_val_prev.items())
-        else:
-            self.evals_d += 1
-
-        if isinstance(x, ooPoint): self._point_id1 = x._id
-        if fixedVarsScheduleID != -1: self._lastDiffVarsID = fixedVarsScheduleID
-
-        derivativeSelf = self._getDerivativeSelf(x, fixedVarsScheduleID, Vars, fixedVars)
-
-        r = Derivative()
-        ac = -1
-        for i, inp in enumerate(self.input):
-            if not isinstance(inp, oofun): continue
-            if inp.discrete: continue
-
-            if inp.is_oovar: 
-                if (Vars is not None and inp not in Vars) or (fixedVars is not None and inp in fixedVars):
-                    continue                
-                ac += 1
-                tmp = derivativeSelf[ac]
-                val = r.get(inp, None)
-                if val is not None:
-                    if isscalar(tmp) or (type(val) == type(tmp) == ndarray and prod(tmp.shape) <= prod(val.shape)): # some sparse matrices has no += implemented 
-                        r[inp] += tmp
-                    else:
-                        if isspmatrix(val) and type(tmp) == DiagonalType:
-                            tmp = tmp.resolve(True)
-                        r[inp] = r[inp] + tmp
-                else:
-                    r[inp] = tmp
-            else:
-                ac += 1
-                
-                elem_d = inp._D(x, fixedVarsScheduleID, Vars=Vars, fixedVars=fixedVars, useSparse = useSparse) 
-                
-                t1 = derivativeSelf[ac]
-                
-                for key, val in elem_d.items():
-                    #if isinstance(t1, Stochastic) or isinstance(val, Stochastic):
-                        #rr = t1 * val
-                    if isinstance(t1, Stochastic) or ((isscalar(val) or isinstance(val, multiarray)) and (isscalar(t1) or isinstance(t1, multiarray))):
-                        rr = t1 * val
-                    elif isinstance(val, Stochastic):
-                        rr = val * t1
-                    elif type(t1) == DiagonalType and type(val) == DiagonalType:
-                        rr = t1 * val
-                    elif type(t1) == DiagonalType or type(val) == DiagonalType:
-                        if isspmatrix(t1): # thus val is DiagonalType
-                            rr = t1._mul_sparse_matrix(val.resolve(True))
-                        else:
-                            if not isPyPy or type(val) != DiagonalType:
-                                rr = t1 *  val #if  type(t1) == DiagonalType or type(val) not in (ndarray, DiagonalType) else (val.T * t1).T   # for PyPy compatibility
-                            else:
-                                rr = (val * t1.T).T
-                    elif isscalar(val) or isscalar(t1) or prod(t1.shape)==1 or prod(val.shape)==1:
-                        rr = (t1 if isscalar(t1) or prod(t1.shape)>1 else asscalar(t1) if isinstance(t1, ndarray) else t1[0, 0]) \
-                        * (val if isscalar(val) or prod(val.shape)>1 else asscalar(val) if isinstance(val, ndarray) else val[0, 0])
-                    else:
-                        if val.ndim < 2: val = atleast_2d(val)
-                        if useSparse is False:
-                            t2 = val
-                        else:
-                            t1, t2 = self._considerSparse(t1, val)
-                        
-                        if not type(t1) == type(t2) ==  ndarray:
-                            # CHECKME: is it trigger somewhere?
-                            if not scipyInstalled:
-                                self.pWarn(scipyAbsentMsg)
-                                rr = dot(t1, t2)
-                            else:
-                                t1 = t1 if isspmatrix_csc(t1) else t1.tocsc() if isspmatrix(t1)  else scipy.sparse.csc_matrix(t1)
-                                t2 = t2 if isspmatrix_csr(t2) else t2.tocsr() if isspmatrix(t2)  else scipy.sparse.csr_matrix(t2)
-                                if t2.shape[0] != t1.shape[1]:
-                                    if t2.shape[1] == t1.shape[1]:
-                                        t2 = t2.T
-                                    else:
-                                        raise FuncDesignerException('incorrect shape in FuncDesigner function _D(), inform developers about the bug')
-                                rr = t1._mul_sparse_matrix(t2)
-                                if useSparse is False:
-                                    rr = rr.toarray() 
-                        else:
-                            rr = dot(t1, t2)
-                    #assert rr.ndim>1
-                    
-                    Val = r.get(key, None)
-                    ValType = type(Val)
-                    if Val is not None:
-                        if type(rr) == DiagonalType:
-                            if ValType == DiagonalType:
-                                
-                                Val = Val + rr # !!!! NOT inplace! (elseware will overwrite stored data used somewhere else)
-                                
-                            else:
-                                tmp  = rr.resolve(useSparse)
-                                if type(tmp) == ndarray and hasattr(Val, 'toarray'):
-                                    Val = Val.toarray()
-                                if type(tmp) == ValType == ndarray and Val.size >= tmp.size:
-                                    Val += tmp
-                                else: # may be problems with sparse matrices inline operation, which are badly done in scipy.sparse for now
-                                    Val = Val + tmp
-                        else:
-                            if isinstance(Val, ndarray) and hasattr(rr, 'toarray'): # i.e. rr is sparse matrix
-                                rr = rr.toarray() # I guess r[key] will hardly be all-zeros
-                            elif hasattr(Val, 'toarray') and isinstance(rr, ndarray): # i.e. Val is sparse matrix
-                                Val = Val.toarray()
-                            if type(rr) == ValType == ndarray and rr.size == Val.size: 
-                                Val += rr
-                            else: 
-                                Val = Val + rr
-                        r[key] = Val
-                    else:
-                        r[key] = rr
-        self._d_val_prev = dict([(key, Copy(value)) for key, value in r.items()])
-        self._d_key_prev = dict([(elem, Copy(x[elem])) for elem in dep]) if involveStore else None
-        return r
-
-    # TODO: handle 2**15 & 0.25 as parameters
-    def _considerSparse(self, t1, t2):  
-        if int64(prod(t1.shape)) * int64(prod(t2.shape)) > 2**15 and   (isinstance(t1, ndarray) and t1.nonzero()[0].size < 0.25*t1.size) or \
-        (isinstance(t2, ndarray) and t2.nonzero()[0].size < 0.25*t2.size):
-            if scipy is None: 
-                self.pWarn(scipyAbsentMsg)
-                return t1,  t2
-            if not isinstance(t1, scipy.sparse.csc_matrix): 
-                t1 = scipy.sparse.csc_matrix(t1)
-            if t1.shape[1] != t2.shape[0]: # can be from flattered t1
-                assert t1.shape[0] == t2.shape[0], 'bug in FuncDesigner Kernel, inform developers'
-                t1 = t1.T
-            if not isinstance(t2, scipy.sparse.csr_matrix): 
-                t2 = scipy.sparse.csr_matrix(t2)
-        return t1,  t2
-
-    def _getDerivativeSelf(self, x, fixedVarsScheduleID, Vars,  fixedVars):
-        Input = self._getInput(x, fixedVarsScheduleID=fixedVarsScheduleID, Vars=Vars,  fixedVars=fixedVars)
-        expectedTotalInputLength = sum([Len(elem) for elem in Input])
-        
-#        if hasattr(self, 'size') and isscalar(self.size): nOutput = self.size
-#        else: nOutput = self(x).size 
-
-        hasUserSuppliedDerivative = self.d is not None
-        if hasUserSuppliedDerivative:
-            derivativeSelf = []
-            if type(self.d) == tuple:
-                if len(self.d) != len(self.input):
-                   raise FuncDesignerException('oofun error: num(derivatives) not equal to neither 1 nor num(inputs)')
-                   
-                for i, deriv in enumerate(self.d):
-                    inp = self.input[i]
-                    if not isinstance(inp, oofun) or inp.discrete: 
-                        #if deriv is not None: 
-                            #raise FuncDesignerException('For an oofun with some input oofuns declared as discrete you have to set oofun.d[i] = None')
-                        continue
-                    
-                    #!!!!!!!!! TODO: handle fixed cases properly!!!!!!!!!!!!
-                    #if hasattr(inp, 'fixed') and inp.fixed: continue
-                    if inp.is_oovar and ((Vars is not None and inp not in Vars) or (fixedVars is not None and inp in fixedVars)):
-                        continue
-                        
-                    if deriv is None:
-                        if not DerApproximatorIsInstalled:
-                            raise FuncDesignerException('To perform gradients check you should have DerApproximator installed, see http://openopt.org/DerApproximator')
-                        derivativeSelf.append(get_d1(self.fun, Input, diffInt=self.diffInt, stencil = self.stencil, \
-                                                     args=self.args, varForDifferentiation = i, pointVal = self._getFuncCalcEngine(x), exactShape = True))
-                    else:
-                        # !!!!!!!!!!!!!! TODO: add check for user-supplied derivative shape
-                        tmp = deriv(*Input)
-                        if not isscalar(tmp) and type(tmp) in (ndarray, tuple, list) and type(tmp) != DiagonalType: # i.e. not a scipy.sparse matrix
-                            tmp = atleast_2d(tmp)
-                            
-                            ########################################
-
-                            _tmp = Input[i]
-                            Tmp = 1 if isscalar(_tmp) or prod(_tmp.shape) == 1 else len(Input[i])
-                            if tmp.shape[1] != Tmp: 
-                                # TODO: add debug msg
-#                                print('incorrect shape in FD AD _getDerivativeSelf')
-#                                print tmp.shape[0], nOutput, tmp
-                                if tmp.shape[0] != Tmp: raise FuncDesignerException('error in getDerivativeSelf()')
-                                tmp = tmp.T
-                                    
-                            ########################################
-
-                        derivativeSelf.append(tmp)
-            else:
-                tmp = self.d(*Input)
-                if not isscalar(tmp) and type(tmp) in (ndarray, tuple, list): # i.e. not a scipy.sparse matrix
-                    tmp = atleast_2d(tmp)
-                    
-                    if tmp.shape[1] != expectedTotalInputLength: 
-                        # TODO: add debug msg
-                        if tmp.shape[0] != expectedTotalInputLength: raise FuncDesignerException('error in getDerivativeSelf()')
-                        tmp = tmp.T
-                        
-                ac = 0
-                if isinstance(tmp, ndarray) and hasattr(tmp, 'toarray') and not isinstance(tmp, multiarray): tmp = tmp.A # is dense matrix
-                
-                #if not isinstance(tmp, ndarray) and not isscalar(tmp) and type(tmp) != DiagonalType:
-                if len(Input) == 1:
-#                    if type(tmp) == DiagonalType: 
-#                            # TODO: mb rework it
-#                            if Input[0].size > 150 and tmp.size > 150:
-#                                tmp = tmp.resolve(True).tocsc()
-#                            else: tmp =  tmp.resolve(False) 
-                    derivativeSelf = [tmp]
-                else:
-                    for i, inp in enumerate(Input):
-                        t = self.input[i]
-                        if t.discrete or (t.is_oovar and ((Vars is not None and t not in Vars) or (fixedVars is not None and t in fixedVars))):
-                            ac += inp.size
-                            continue                                    
-                        if isinstance(tmp, ndarray):
-                            TMP = tmp[:, ac:ac+Len(inp)]
-                        elif isscalar(tmp):
-                            TMP = tmp
-                        elif type(tmp) == DiagonalType: 
-                            if tmp.size == inp.size and ac == 0:
-                                TMP = tmp
-                            else:
-                                # print debug warning here
-                                # TODO: mb rework it
-                                if inp.size > 150 and tmp.size > 150:
-                                    tmp = tmp.resolve(True).tocsc()
-                                else: tmp =  tmp.resolve(False) 
-                                TMP = tmp[:, ac:ac+inp.size]
-                        else: # scipy.sparse matrix
-                            TMP = tmp.tocsc()[:, ac:ac+inp.size]
-                        ac += Len(inp)
-                        derivativeSelf.append(TMP)
-                    
-            # TODO: is it required?
-#                if not hasattr(self, 'outputTotalLength'): self(x)
-#                
-#                if derivativeSelf.shape != (self.outputTotalLength, self.inputTotalLength):
-#                    s = 'incorrect shape for user-supplied derivative of oofun '+self.name+': '
-#                    s += '(%d, %d) expected, (%d, %d) obtained' % (self.outputTotalLength, self.inputTotalLength,  derivativeSelf.shape[0], derivativeSelf.shape[1])
-#                    raise FuncDesignerException(s)
-        else:
-            if Vars is not None or fixedVars is not None: raise FuncDesignerException("sorry, custom oofun derivatives don't work with Vars/fixedVars arguments yet")
-            if not DerApproximatorIsInstalled:
-                raise FuncDesignerException('To perform this operation you should have DerApproximator installed, see http://openopt.org/DerApproximator')
-                
-            derivativeSelf = get_d1(self.fun, Input, diffInt=self.diffInt, stencil = self.stencil, args=self.args, pointVal = self._getFuncCalcEngine(x), exactShape = True)
-            if type(derivativeSelf) == tuple:
-                derivativeSelf = list(derivativeSelf)
-            elif type(derivativeSelf) != list:
-                derivativeSelf = [derivativeSelf]
-        
-        #assert all([elem.ndim > 1 for elem in derivativeSelf])
-       # assert len(derivativeSelf[0])!=16
-        #assert (type(derivativeSelf[0]) in (int, float)) or derivativeSelf[0][0]>480.00006752 or derivativeSelf[0][0]<480.00006750
-        return derivativeSelf
+    _getDerivativeSelf = getDerivativeSelf
+    
+    ##########################
+    # !! May be overloaded by some oofuns,
+    # thus cannot be moved from here
+    _D = _D
+    ##########################
 
     def D2(self, x):
-        raise FuncDesignerException('2nd derivatives for obj-funcs are not implemented yet')
+        raise FuncDesignerException('2nd derivatives for oofuns are not implemented yet')
 
     def check_d1(self, point):
         if self.d is None:
@@ -1866,11 +1361,6 @@ class oofun(OOFun):
     
     """                                             End of class oofun                                             """
 
-        
-class Derivative(dict):
-    def __init__(self):
-        pass
-
 
 def atleast_oofun(arg):
     if isinstance(arg, oofun):
@@ -1884,14 +1374,3 @@ def atleast_oofun(arg):
     else:
         #return oofun(lambda *args, **kwargs: arg(*args,  **kwargs), input=None, discrete=True)
         raise FuncDesignerException('incorrect type for the function _atleast_oofun')
-
-def mul_aux_d(x, y):
-    Xsize, Ysize = Len(x), Len(y)
-    if Xsize == 1:
-        return Copy(y)
-    elif Ysize == 1:
-        return Diag(None, scalarMultiplier = y, size = Xsize)
-    elif Xsize == Ysize:
-        return Diag(y)
-    else:
-        raise FuncDesignerException('for oofun multiplication a*b should be size(a)=size(b) or size(a)=1 or size(b)=1')   
